"""
kombu JSON-RPC server integration.
"""

import logging

import kombu.mixins

import pjrpc

logger = logging.getLogger(__package__)


class Executor(kombu.mixins.ConsumerProducerMixin):
    """
    `kombu <http://kombu.readthedocs.org/>`_ based JSON-RPC server.

    :param broker_url: broker connection url
    :param queue_name: requests queue name
    :param conn_args: additional connection arguments
    :param queue_args: queue arguments
    :param publish_args: message publish additional arguments
    :param prefetch_count: worker prefetch count
    :param kwargs: dispatcher additional arguments
    """

    def __init__(
        self, broker_url, queue_name, conn_args=None, queue_args=None, publish_args=None, prefetch_count=0, **kwargs
    ):
        self.connection = kombu.Connection(broker_url, **(conn_args or {}))

        self._rpc_queue = kombu.Queue(queue_name, **(queue_args or {}))
        self._prefetch_count = prefetch_count
        self._publish_args = publish_args

        self._dispatcher = pjrpc.server.Dispatcher(**kwargs)

    @property
    def dispatcher(self):
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    def get_consumers(self, Consumer, channel):
        return [Consumer(
            queues=[self._rpc_queue],
            on_message=self._rpc_handle,
            accept={'application/json'},
            prefetch_count=self._prefetch_count,
        )]

    def _rpc_handle(self, message):
        """
        Handles JSON-RPC request.

        :param message: kombu message :py:class:`kombu.message.Message`
        """

        try:
            reply_to = message.properties.get('reply_to')
            response_text = self._dispatcher.dispatch(message.body, context=message)

            if response_text is not None:
                if reply_to is None:
                    logger.warning("property 'reply_to' is missing")
                else:
                    self.producer.publish(
                        response_text,
                        routing_key=reply_to,
                        correlation_id=message.properties.get('correlation_id'),
                        content_type='application/json',
                        content_encoding='utf8',
                        **(self._publish_args or {})
                    )

            message.ack()

        except Exception as e:
            logger.exception("jsonrpc request handling error: %s", e)



"""
JSON-RPC server package.
"""

from .dispatcher import Dispatcher, AsyncDispatcher, Method, MethodRegistry, View

__all__ = [
    'Dispatcher',
    'AsyncDispatcher',
    'Method',
    'MethodRegistry',
    'View',
]



import pytest
from pjrpc.server import validators
from pjrpc.server.validators import jsonschema


@pytest.mark.parametrize('dyn_method, params, schema', [
    (
        'param1, param2',
        [1, 2],
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
            }
        }
    ),
    (
        'param1, *args',
        [1, 2, 3],
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'args': {
                    'type': 'array',
                    'items': {
                        'type': 'integer'
                    }
                },
            }
        }
    ),
    (
        'param1, *args',
        [1],
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'args': {
                    'type': 'array',
                    'items': {
                        'type': 'integer'
                    }
                },
            }
        }
    ),
    (
        'param1=1',
        [],
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
            }
        }

    ),
    (
        'param1, param2=2',
        [1],
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
            }
        }
    ),
    (
        'param1, param2',
        {
            'param1': 1,
            'param2': 2
        },
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
            }
        }
    ),
    (
        'param1, **kwargs',
        {
            'param1': 1,
            'param2': 2,
            'param3': 3,
        },
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'kwargs': {'type': 'object'},
            }
        }
    ),
    (
        'param1, *, param2, param3',
        {
            'param1': 1,
            'param2': 2,
            'param3': 3
        },
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
                'param3': {'type': 'integer'},
            }
        }

    ),
    (
        'param1=1',
        {},
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
            }
        }
    ),
    (
        'param1, param2=2',
        {
            'param1': 1
        },
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
            }
        }
    ),
    (
        'param1, *, param2=2',
        {
            'param1': 1
        },
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
            }
        }
    ),
], indirect=['dyn_method'])
def test_validation_success(dyn_method, params, schema):
    validator = jsonschema.JsonSchemaValidator()
    validator.validate_method(dyn_method, params, schema=schema)


@pytest.mark.parametrize('dyn_method, params, schema', [
    (
        'param1, param2',
        [1],
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
            }
        }
    ),
    (
        'param1, param2',
        [1, 2, 3],
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
            }
        }
    ),
    (
        'param1, *args',
        [],
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'args': {
                    'type': 'array',
                    'items': {
                        'type': 'integer'
                    }
                },
            }
        }
    ),
    (
        'param1, param2',
        {
            'param1': 1
        },
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
            }
        }
    ),
    (
        'param1, param2',
        {
            'param1': 1,
            'param2': 2,
            'param3': 3
        },
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
            }
        }
    ),
    (
        'param1, **kwargs',
        {
            'param2': 2
        },
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'kwargs': {'type': 'object'},
            }
        }
    ),
    (
        'param1, *, param2, param3',
        {
            'param2': 1
        },
        {
            'type': 'object',
            'properties': {
                'param1': {'type': 'integer'},
                'param2': {'type': 'integer'},
                'param3': {'type': 'integer'},
            }
        }
    ),
], indirect=['dyn_method'])
def test_validation_error(dyn_method, params, schema):
    validator = jsonschema.JsonSchemaValidator()

    with pytest.raises(validators.ValidationError):
        validator.validate_method(dyn_method, params, schema=schema)



import json

import pjrpc


class Unset:
    """
    `Sentinel <https://en.wikipedia.org/wiki/Sentinel_value>`_ object.
    Used to distinct unset (missing) values from ``None`` ones.
    """

    def __bool__(self):
        return False

    def __repr__(self):
        return "UNSET"

    def __str__(self):
        return repr(self)


UNSET = Unset()


class JSONEncoder(json.JSONEncoder):
    """
    Library default JSON encoder. Encodes request, response and error objects to be json serializable.
    All custom encoders should be inherited from it.
    """

    def default(self, o):
        if isinstance(
            o, (pjrpc.Response, pjrpc.Request, pjrpc.BatchResponse, pjrpc.BatchRequest, pjrpc.exceptions.JsonRpcError)
        ):
            return o.to_json()

        return super().default(o)



"""
aiohttp JSON-RPC server integration.
"""

from aiohttp import web

import pjrpc


class Application(web.Application):
    """
    `aiohttp <https://aiohttp.readthedocs.io/en/stable/web.html>`_ based JSON-RPC server.

    :param path: JSON-RPC handler base path
    :param app_args: arguments to be passed to :py:class:`aiohttp.web.Application`
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.AsyncDispatcher`
    """

    def __init__(self, path='', app_args=None, **kwargs):
        super().__init__(**(app_args or {}))
        self._dispatcher = pjrpc.server.AsyncDispatcher(**kwargs)

        self.router.add_post(path, self.rpc_handle)

    @property
    def dispatcher(self):
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    async def rpc_handle(self, http_request):
        """
        Handles JSON-RPC request.

        :param http_request: :py:class:`aiohttp.web.Response`
        :returns: :py:class:`aiohttp.web.Request`
        """

        if http_request.content_type != 'application/json':
            raise web.HTTPUnsupportedMediaType()

        try:
            request_text = await http_request.text()
        except UnicodeDecodeError as e:
            raise web.HTTPBadRequest() from e

        response_text = await self._dispatcher.dispatch(request_text, context=http_request)
        if response_text is None:
            return web.Response()
        else:
            return web.json_response(text=response_text)



import asyncio
import collections
import functools as ft
import json
import logging

import pjrpc
from pjrpc.common import v20, UNSET
from . import validators

logger = logging.getLogger(__package__)

default_validator = validators.base.BaseValidator()


class Method:
    """
    JSON-RPC method wrapper. Stores method itself and some metainformation.

    :param method: method
    :param name: method name
    :param context: context name
    """

    def __init__(self, method, name=None, context=None):
        self.method = method
        self.name = name or getattr(method, 'name', method.__name__)
        self.context = context
        self.validator, self.validator_args = getattr(method, '__pjrpc_meta__', (default_validator, {}))

    def bind(self, params, context=None):
        method_params = self.validator.validate_method(
            self.method, params, exclude=(self.context,) if self.context else (), **self.validator_args
        )

        if self.context is not None:
            method_params[self.context] = context

        return ft.partial(self.method, **method_params)


class ViewMethod(Method):
    """
    View method.

    :param view_cls: view class
    :param name: view class method name
    :param context: context name
    """

    def __init__(self, view_cls, name, context=None):
        super().__init__(view_cls, name, context)

        self.view_cls = view_cls

    def bind(self, params, context=None):
        view = self.view_cls(context) if self.context else self.view_cls()
        method = getattr(view, self.name)

        method_params = self.validator.validate_method(method, params, **self.validator_args)

        return ft.partial(method, **method_params)


class View:
    """
    Class based method handler.
    """

    @classmethod
    def __methods__(cls):
        for attr_name in filter(lambda name: not name.startswith('_'), dir(cls)):
            attr = getattr(cls, attr_name)
            if callable(attr):
                yield attr


class MethodRegistry:
    """
    Method registry.

    :param prefix: method name prefix to be used for naming containing methods
    """

    def __init__(self, prefix=None):
        self._prefix = prefix
        self._registry = {}

    def __iter__(self):
        """
        Returns registry method iterator.
        """

        return iter(self._registry)

    def __getitem__(self, item):
        """
        Returns a method from the registry by name.

        :param item: method name
        :returns: found method
        :raises: KeyError
        """

        return self._registry[item]

    def get(self, item):
        """
        Returns a method from the registry by name.

        :param item: method name
        :returns: found method or `None`
        """

        return self._registry.get(item)

    def add(self, maybe_method=None, name=None, context=None):
        """
        Decorator adding decorated method to the registry.

        :param maybe_method: method or `None`
        :param name: method name to be used instead of `__name__` attribute
        :param context: parameter name to be used as an application context
        :returns: decorated method or decorator
        """

        def decorator(method):
            full_name = '.'.join(filter(None, (self._prefix, name or getattr(method, 'name', method.__name__))))
            self.add_methods(Method(method, full_name, context))

            return method

        if maybe_method is None:
            return decorator
        else:
            return decorator(maybe_method)

    def add_methods(self, *methods):
        """
        Adds methods to the registry.

        :param methods: methods to be added. Each one can be an instance of :py:class:`pjrpc.server.Method`
                        or plain method
        """

        for method in methods:
            if isinstance(method, Method):
                self._add_method(method.name, method)
            else:
                self.add(method)

    def view(self, maybe_view=None, context=None, prefix=None):
        """
        Methods view decorator.

        :param maybe_view: view class instance or `None`
        :param context: application context name
        :param prefix: view methods prefix
        :return: decorator or decorated view
        """

        def decorator(view):
            for method in view.__methods__():
                full_name = '.'.join(filter(None, (self._prefix, prefix, method.__name__)))
                self._add_method(full_name, ViewMethod(view, method.__name__, context))

            return view

        # maybe_view's type depends on the usage of the decorator.  It's a View
        # if it's used as `@view` but ``None`` if used as `@view()`.
        if maybe_view is None:
            return decorator
        else:
            return decorator(maybe_view)

    def merge(self, other):
        """
        Merges two registries.

        :param other: registry to be merged in the current one
        """

        for name in other:
            self._add_method(name, other[name])

    def _add_method(self, name, method):
        if name in self._registry:
            logger.warning(f"method '{name}' already registered")

        self._registry[name] = method


class Dispatcher:
    """
    Method dispatcher.

    :param request_class: JSON-RPC request class
    :param response_class: JSON-RPC response class
    :param batch_request: JSON-RPC batch request class
    :param batch_response: JSON-RPC batch response class
    :param json_loader: request json loader
    :param json_dumper: response json dumper
    :param json_encoder: response json encoder
    :param json_decoder: request json decoder
    :param error_handler: error handling function
    """

    def __init__(
        self,
        *,
        request_class=v20.Request,
        response_class=v20.Response,
        batch_request=v20.BatchRequest,
        batch_response=v20.BatchResponse,
        json_loader=json.loads,
        json_dumper=json.dumps,
        json_encoder=None,
        json_decoder=None,
        error_handler=None,
    ):
        self._json_loader = json_loader
        self._json_dumper = json_dumper
        self._json_encoder = json_encoder or pjrpc.JSONEncoder
        self._json_decoder = json_decoder
        self._request_class = request_class
        self._response_class = response_class
        self._batch_request = batch_request
        self._batch_response = batch_response
        self._error_handler = error_handler or self._handle_jsonrpc_error

        self._registry = MethodRegistry()

    def add(self, method, name=None, context=None):
        """
        Adds method to the registry.

        :param method: method
        :param name: method name
        :param context: application context name
        :returns:
        """

        self._registry.add(method, name, context)

    def add_methods(self, *methods):
        """
        Adds methods to the registry.

        :param methods: method list. Each method may be an instance of :py:class:`pjrpc.server.MethodRegistry`,
                        :py:class:`pjrpc.server.Method` or plain function
        """

        for method in methods:
            if isinstance(method, MethodRegistry):
                self._registry.merge(method)
            elif isinstance(method, Method):
                self._registry.add_methods(method)
            else:
                self._registry.add(method)

    def view(self, view):
        """
        Adds class based view to the registry.

        :param view: view to be added
        """

        self._registry.view(view)

    def dispatch(self, request_text, context=None):
        """
        Deserializes request, dispatches it to the required method and serializes the result.

        :param request_text: request text representation
        :param context: application context (if supported)
        :return: response text representation
        """

        logger.debug("request received: %s", request_text)

        try:
            request_json = self._json_loader(request_text, cls=self._json_decoder)
            if isinstance(request_json, (list, tuple)):
                request = self._batch_request.from_json(request_json)
            else:
                request = self._request_class.from_json(request_json)

        except json.JSONDecodeError as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.ParseError(data=str(e)))

        except (pjrpc.exceptions.DeserializationError, pjrpc.exceptions.IdentityError) as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.InvalidRequestError(data=str(e)))

        else:
            if isinstance(request, collections.Iterable):
                response = self._batch_response(*filter(lambda resp: resp is not UNSET, (
                    self._handle_rpc_request(request, context) for request in request
                )))

            else:
                response = self._handle_rpc_request(request, context)

        if response is not UNSET:
            response_text = self._json_dumper(response.to_json(), cls=self._json_encoder)
            logger.debug("response sent: %s", response_text)

            return response_text

    def _handle_rpc_request(self, request, context):
        response_id, result, error = request.id, UNSET, UNSET

        try:
            result = self._handle_rpc_method(request.method, request.params, context)
        except pjrpc.exceptions.JsonRpcError as e:
            logger.info("method execution error %s(%r): %r", request.method, request.params, e)
            error = e
        except Exception as e:
            logger.exception("internal server error: %r", e)
            error = pjrpc.exceptions.InternalError()

        if response_id is None:
            return UNSET

        if error:
            error = self._error_handler(error)

        return self._response_class(id=response_id, result=result, error=error)

    def _handle_rpc_method(self, method_name, params, context):
        method = self._registry.get(method_name)
        if method is None:
            raise pjrpc.exceptions.MethodNotFoundError(data=f"method '{method_name}' not found")

        try:
            method = method.bind(params, context=context)
        except validators.ValidationError as e:
            raise pjrpc.exceptions.InvalidParamsError(data=e) from e

        try:
            return method()

        except pjrpc.exceptions.JsonRpcError:
            raise

        except Exception as e:
            logger.exception("method unhandled exception %s(%r): %r", method_name, params, e)
            raise pjrpc.exceptions.ServerError() from e

    def _handle_jsonrpc_error(self, error):
        if error and error.data:
            error.data = str(error.data)

        return error


class AsyncDispatcher(Dispatcher):
    """
    Asynchronous method dispatcher.
    """

    async def dispatch(self, request_text, context=None):
        """
        Deserializes request, dispatches it to the required method and serializes the result.

        :param request_text: request text representation
        :param context: application context (if supported)
        :return: response text representation
        """

        logger.debug("request received: %s", request_text)

        try:
            request_json = self._json_loader(request_text, cls=self._json_decoder)
            if isinstance(request_json, (list, tuple)):
                request = self._batch_request.from_json(request_json)
            else:
                request = self._request_class.from_json(request_json)

        except json.JSONDecodeError as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.ParseError(data=str(e)))

        except (pjrpc.exceptions.DeserializationError, pjrpc.exceptions.IdentityError) as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.InvalidRequestError(data=str(e)))

        else:
            if isinstance(request, collections.Iterable):
                response = self._batch_response(*filter(lambda resp: resp is not UNSET, await asyncio.gather(
                    *(self._handle_rpc_request(request, context) for request in request)
                )))

            else:
                response = await self._handle_rpc_request(request, context)

        if response is not UNSET:
            response_text = self._json_dumper(response.to_json(), cls=self._json_encoder)
            logger.debug("response sent: %s", response_text)

            return response_text

    async def _handle_rpc_request(self, request, context):
        response_id, result, error = request.id, UNSET, UNSET

        try:
            result = await self._handle_rpc_method(request.method, request.params, context)
        except pjrpc.exceptions.JsonRpcError as e:
            logger.info("method execution error %s(%r): %r", request.method, request.params, e)
            error = e
        except Exception as e:
            logger.exception("internal server error: %r", e)
            error = pjrpc.exceptions.InternalError()

        if response_id is None:
            return UNSET

        if error:
            error = self._error_handler(error)

        return self._response_class(id=response_id, result=result, error=error)

    async def _handle_rpc_method(self, method_name, params, context):
        method = self._registry.get(method_name)
        if method is None:
            raise pjrpc.exceptions.MethodNotFoundError(data=f"method '{method_name}' not found")

        try:
            method = method.bind(params, context=context)
        except validators.ValidationError as e:
            raise pjrpc.exceptions.InvalidParamsError(data=e) from e

        try:
            result = method()
            if asyncio.iscoroutine(result):
                result = await result

            return result

        except pjrpc.exceptions.JsonRpcError:
            raise

        except Exception as e:
            logger.exception("method unhandled exception %s(%r): %r", method_name, params, e)
            raise pjrpc.exceptions.ServerError() from e



import jsonschema

from . import base


class JsonSchemaValidator(base.BaseValidator):
    """
    Parameters validator based on `jsonschema <https://python-jsonschema.readthedocs.io/en/stable/>`_ library.

    :param kwargs: default jsonschema validator arguments
    """

    def __init__(self, **kwargs):
        kwargs.setdefault('types', {'array': (list, tuple)})
        self.default_kwargs = kwargs

    def validate_method(self, method, params, exclude=(), **kwargs):
        """
        Validates params against method using ``pydantic`` validator.

        :param method: method to validate parameters against
        :param params: parameters to be validated
        :param exclude: parameter names to be excluded from validation
        :param kwargs: jsonschema validator arguments

        :raises: :py:class:`pjrpc.server.validators.ValidationError`
        """

        arguments = super().validate_method(method, params, exclude)

        try:
            kwargs = {**self.default_kwargs, **kwargs}
            jsonschema.validate(arguments, **kwargs)
        except jsonschema.ValidationError as e:
            raise base.ValidationError(e) from e

        return arguments
